/*

Copyright (C) 2003 Christian Kreibich <christian@whoop.org>.

Permission is hereby granted, free of charge, to any person obtaining a copy
of this software and associated documentation files (the "Software"), to
deal in the Software without restriction, including without limitation the
rights to use, copy, modify, merge, publish, distribute, sublicense, and/or
sell copies of the Software, and to permit persons to whom the Software is
furnished to do so, subject to the following conditions:

The above copyright notice and this permission notice shall be included in
all copies of the Software and its documentation and acknowledgment shall be
given in the documentation and software packages that this Software was
used.

THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS OR
IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL
THE AUTHORS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER LIABILITY, WHETHER
IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING FROM, OUT OF OR IN
CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE SOFTWARE.

*/

#define __FAVOR_BSD

#include <unistd.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/ip.h>
#include <getopt.h>

#ifndef MAXPATHLEN
#define MAXPATHLEN 4096
#endif

#include <pcap.h>
#include <libnet.h>

#define TCP_FIRE_LINK 0
#define TCP_FIRE_RAW  1

struct pcap_cb_data
{
  int            mode;        /* whether we're writing raw IP or linklevel data */
  u_int          loops;       /* how often to write this out */

  char          *tracefile;   /* input filename */
  char          *iface;       /* interface name, default eth0, as given by user */

  pcap_t        *pcap;
  libnet_t      *ln;

  u_int          total;
  u_int          written;
};


static char pcap_errbuf[PCAP_ERRBUF_SIZE];
static char libnet_errbuf[LIBNET_ERRBUF_SIZE];


static void 
tcpf_pcap_cb(u_char *user,
	     const struct pcap_pkthdr *pkthdr,
	     const u_char *packet)
{
  struct pcap_cb_data *data = (struct pcap_cb_data *) user;

  data->total += pkthdr->caplen;

  switch (data->mode)
    {
    case TCP_FIRE_RAW:
      data->written += libnet_write_raw_ipv4(data->ln, (u_char *) packet, pkthdr->caplen);
      break;
    case TCP_FIRE_LINK:
    default:
      data->written += libnet_write_link(data->ln, (u_char *) packet, pkthdr->caplen);
    }
}


static void
tcpf_print_usage_exit(const char *progname)
{
  printf("USAGE: %s [-h] [-i <interface>] [-l <loops>] <input tracefile>\n\n"
	 "Takes a tcpdump trace file and writes its packets to the interface\n"
	 "passed with the -i option. Default interface is eth0. The file can\n"
	 "be written repeatedly, using the -l options. Default is once. If you\n"
	 "pass -l 0, looping is indefinitely.\n",
	 progname);
  
  exit(0);
}


int 
main(int argc, char** argv)
{
  u_int  i;
  char opt;
  struct pcap_cb_data cb_data;

  if (argc < 2)
    tcpf_print_usage_exit(argv[0]);	  

  /* Initialize our data container */
  memset(&cb_data, 0, sizeof(struct pcap_cb_data));
  cb_data.iface = "eth0";
  cb_data.loops = 1;
  
  while ( (opt = getopt(argc, argv, "hi:l:")) != EOF)
    {
      switch (opt)
	{
	case 'h':
	  tcpf_print_usage_exit(argv[0]);
	  break;
	case 'i':
	  cb_data.iface = optarg;
	  break;
	case 'l':
	  cb_data.loops = atoi(optarg);
	  break;
	default:
	  tcpf_print_usage_exit(argv[0]);	  
	}
    }

  cb_data.tracefile = argv[optind];

  /* If loop count is zero, then make it really really big */
  if (cb_data.loops == 0)
    cb_data.loops = (u_int) -1;

  /* Open the input trace file for reading: */

  if ((cb_data.pcap = pcap_open_offline(cb_data.tracefile, pcap_errbuf)) == NULL)
    {
      printf("Could not open source trace file %s\n", cb_data.tracefile);
      exit(-1);
    }

  printf("Opening connection to '%s'\n", cb_data.iface);

  switch(pcap_datalink(cb_data.pcap))
    {
    case DLT_RAW:
    case DLT_NULL:
      cb_data.mode = TCP_FIRE_RAW;
      cb_data.ln = libnet_init(LIBNET_RAW4, cb_data.iface, libnet_errbuf);
      break;
    case DLT_EN10MB:
    case DLT_EN3MB:
      cb_data.mode = TCP_FIRE_LINK;
      cb_data.ln = libnet_init(LIBNET_LINK, cb_data.iface, libnet_errbuf);
      break;
    default:
      printf("This tool is only supposed to be used on traces containing\n"
	     "link-level headers or raw IP data. Sorry ..\n");
      exit(-1);
    }

  if (!cb_data.ln)
    {
      printf("Could not create libnet handle: %s", libnet_errbuf);
      pcap_close(cb_data.pcap);  
      exit(-1);
    }

  /* Write the packets out in a callback loop */
  for (i = 0; i < cb_data.loops; i++)
    {
      printf("%i. round ...\n", i);
      pcap_loop(cb_data.pcap, -1, tcpf_pcap_cb, (u_char*) &cb_data);
      
      if (fseek(pcap_file(cb_data.pcap), sizeof(struct pcap_file_header), SEEK_SET) < 0)
	{
	  printf("Looping writes failed, quitting. Euan, please\n"
		 "do not kill Christian now :)\n");
	}
    }
  
  printf("%u bytes input, %u written, %u bytes dropped.\n",
	 cb_data.total, cb_data.written, cb_data.total - cb_data.written);
  
  pcap_close(cb_data.pcap);  
  libnet_destroy(cb_data.ln);

  /* Mask any error codes pcap_loop may have left over. */
  exit(0);
}
